# -*- coding: utf-8 -*-
"""
Utilities for parsing function signatures to make easy CLI's or to call
functions programmatically
"""
import inspect
from inspect import signature
import logging

logger = logging.getLogger(__name__)


def arg_to_str(arg):
    """Format input as str w/ appropriate quote types for python call

    Returns
    --------
    out : str
        String rep of the input arg with proper quotation for formatting in
        python -c '{commands}' cli calls. For example, int, float, None -> '0'
        str -> \"string\"
    """
    if isinstance(arg, str):
        return f'"{arg}"'
    else:
        return f'{arg}'


def has_class(obj):
    """Determine whether an object is a method that is bound to a class

    Returns
    -------
    out : bool
        Whether the input object belongs to a class. For example,
        MyClass.bound_method will return True, whereas standalone_fun will
        return False.
    """
    if hasattr(obj, '__qualname__'):
        if len(obj.__qualname__.split('.')) > 1:
            return True
    else:
        return False


def get_class(obj):
    """Get the name of the class that the method object is bound to. Returns an
    empty string if the method object is not bound to a class.

    Returns
    -------
    out : str
        The class name the input object method belongs to. For example,
        MyClass.bound_method will return "MyClass", whereas standalone_fun will
        return an empty string.
    """
    class_name = ''
    if has_class(obj):
        class_name = obj.__qualname__.split('.')[0]

    return class_name


def is_standalone_fun(obj):
    """Determine whether an object is a standalone function without a class

    Returns
    -------
    out : bool
        Whether the input object is a standalone function. For example,
        MyClass.bound_method will return False, whereas standalone_fun will
        return True.
    """
    return inspect.isfunction(obj) and not has_class(obj)


def get_fun_str(fun):
    """Get the function string from a function object including the
    ClassName.function if the function is bound

    Returns
    -------
    out : str
        The function string to call the input function. For example
        MyClass.bound_method will return "MyClass.bound_method", whereas
        standalone_fun will return "standalone_fun".
    """
    fun_name = fun.__name__
    if is_standalone_fun(fun):
        return fun_name
    elif has_class(fun):
        class_name = get_class(fun)
        return f'{class_name}.{fun_name}'
    else:
        return fun_name


def get_arg_str(fun, config):
    """Get a string representation of positional and keyword arguments required
    by an input function and provided in the config dictionary.

    Example
    -------
    If the function signature is my_fun(a, b, c=0) and config is
    {'a': 1, 'b': 2, 'c': 3}, the returned arg_str will be "1, 2, c=3". The
    function can also take *args or **kwargs, which will be taken from the
    "args" and "kwargs" keys in the config. "args" must be mapped to a list,
    and "kwargs" must be mapped to a dictionary.

    Parameters
    ----------
    fun : obj
        A callable object with a function signature. The function signature
        will be parsed for args and kwargs which will be taken from the config.
    config : dict
        A namespace of arguments to run fun. Not all entries in config may be
        used, but all required inputs to fun must be provided in config. Can
        include "args" and "kwargs" which must be mapped to a list and a
        dictionary, respectively.

    Returns
    -------
    arg_str : str
        Argument string that can be used to call fun programmatically, e.g.
        fun(arg_str)
    """

    sig = signature(fun)

    arg_strs = []

    for arg_name, value in sig.parameters.items():
        is_self = arg_name == 'self'
        is_kw = value.default != value.empty
        is_star_arg = str(value).startswith('*') and str(value).count('*') == 1
        is_star_kwa = str(value).startswith('*') and str(value).count('*') == 2
        not_required = (is_self or is_kw or is_star_arg or is_star_kwa)
        required = not not_required

        if arg_name in config:
            if not is_kw and not (is_star_arg or is_star_kwa):
                arg = arg_to_str(config[arg_name])
                arg_strs.append(f'{arg}')

            elif is_kw and not (is_star_arg or is_star_kwa):
                arg = arg_to_str(config[arg_name])
                arg_strs.append(f'{arg_name}={arg}')

            elif is_star_arg:
                msg = '"args" key in config must be mapped to a list!'
                assert isinstance(config[arg_name], (list, tuple)), msg
                arg_strs += [f'{arg_to_str(star_arg)}'
                             for star_arg in config[arg_name]]

            elif is_star_kwa:
                msg = '"kwargs" key in config must be mapped to a dict!'
                assert isinstance(config[arg_name], dict), msg
                arg_strs += [f'{star_name}={arg_to_str(star_kw)}'
                             for star_name, star_kw
                             in config[arg_name].items()]

        elif required:
            msg = (f'Positional argument "{arg_name}" '
                   'needs to be defined in config!')
            logger.error(msg)
            raise KeyError(msg)

    arg_str = ', '.join(arg_strs)

    return arg_str


def get_fun_call_str(fun, config, quote_char='\"'):
    """Get a string that will call a function using args and kwargs from a
    generic config.

    Example
    -------
    If the function signature is my_fun(a, b, c=0) and config is
    {'a': 1, 'b': 2, 'c': 3}, the returned call string will be
    "my_fun(1, 2, c=3)". The function can also take *args or **kwargs, which
    will be taken from the "args" and "kwargs" keys in the config. "args" must
    be mapped to a list, and "kwargs" must be mapped to a dictionary.

    Parameters
    ----------
    fun : obj
        A callable object with a function signature. The function signature
        will be parsed for args and kwargs which will be taken from the config.
    config : dict
        A namespace of arguments to run fun. Not all entries in config may be
        used, but all required inputs to fun must be provided in config. Can
        include "args" and "kwargs" which must be mapped to a list and a
        dictionary, respectively.
    quote_char : str
        Character to use for string quotes in the fun_call_str output.

    Returns
    -------
    fun_call_str : str
        A string representation of a function call e.g. "fun(arg1, arg2,
        kw1=kw1)" where arg1, arg2, and kw1 were found in the config.
    """

    fun_str = get_fun_str(fun)
    arg_str = get_arg_str(fun, config)
    call_str = f'{fun_str}({arg_str})'

    if quote_char:
        call_str = call_str.replace('"', quote_char)
        call_str = call_str.replace("'", quote_char)
        call_str = call_str.replace("/'", quote_char)
        call_str = call_str.replace('/"', quote_char)

    return call_str



# -*- coding: utf-8 -*-
"""
General Joint Probabilty Distribution calculator
"""
from concurrent.futures import as_completed
import gc
import logging
import h5py
import numpy as np
import os
import pandas as pd
from warnings import warn

from rex.version import __version__
from rex.renewable_resource import WindResource
from rex.resource import Resource
from rex.utilities.execution import SpawnProcessPool
from rex.utilities.loggers import log_mem, log_versions
from rex.utilities.utilities import slice_sites, to_records_array

logger = logging.getLogger(__name__)


class JointPD:
    """
    Compute the joint probability distribution between the desired variables
    """
    def __init__(self, res_h5, res_cls=Resource, hsds=False):
        """
        Parameters
        ----------
        res_h5 : str
            Path to resource h5 file(s)
        res_cls : Class, optional
            Resource handler class to use to access res_h5,
            by default Resource
        hsds : bool, optional
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS, by default False
        """
        log_versions(logger)
        self._res_h5 = res_h5
        self._res_cls = res_cls
        self._hsds = hsds

    @property
    def res_h5(self):
        """
        Path to resource h5 file(s)

        Returns
        -------
        str
        """
        return self._res_h5

    @property
    def res_cls(self):
        """
        Resource class to use to access wind_h5

        Returns
        -------
        Class
        """
        return self._res_cls

    @staticmethod
    def _make_bins(start, stop, step):
        """
        Create bin edges from bin range

        Parameters
        ----------
        start : int
            bin range start value
        stop : int
            bin range stop value
        step : int
            bin range step value

        Returns
        -------
        bin_edges : ndarray
            Vector of inclusive bin edges
        """
        bin_edges = np.arange(start, stop + step, step)

        return bin_edges

    @classmethod
    def compute_joint_pd(cls, res_h5, dset1, dset2, bins1, bins2,
                         res_cls=Resource, hsds=False,
                         sites_slice=None):
        """
        Compute the joint probability distribution between the two given
        datasets using the given bins for given sites

        Parameters
        ----------
        res_h5 : str
            Path to resource h5 file(s)
        dset1 : str
            Dataset 1 to generate joint probability distribution for
        dset2 : str
            Dataset 2 to generate joint probabilty distribution for
        bins1 : tuple
            (start, stop, step) for dataset 1 bins. The stop value is
            inclusive, so (0, 6, 2) would yield three bins with edges (0, 2, 4,
            6). If the stop value is not perfectly divisible by the step, the
            last bin will overshoot the stop value.
        bins2 : tuple
            (start, stop, step) for dataset 2 bins. The stop value is
            inclusive, so (0, 6, 2) would yield three bins with edges (0, 2, 4,
            6). If the stop value is not perfectly divisible by the step, the
            last bin will overshoot the stop value.
        sites : list | slice, optional
        res_cls : Class, optional
            Resource handler class to use to access res_h5,
            by default Resource
        hsds : bool, optional
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS, by default False
        sites_slice : slice | None, optional
            Sites to extract, if None all, by default None

        Returns
        -------
        jpd : dict
            Dictionary of joint probabilty distribution densities for given
            sites
        """
        if sites_slice is None:
            sites_slice = slice(None, None, None)
        elif isinstance(sites_slice, int):
            sites_slice = [sites_slice]

        with res_cls(res_h5, hsds=hsds) as f:
            arr1 = f[dset1, :, sites_slice]
            arr2 = f[dset2, :, sites_slice]

        bins1 = cls._make_bins(*bins1)
        bins2 = cls._make_bins(*bins2)

        if isinstance(sites_slice, slice) and sites_slice.stop:
            gids = list(range(*sites_slice.indices(sites_slice.stop)))
        elif isinstance(sites_slice, (list, np.ndarray)):
            gids = sites_slice

        jpd = {}
        for i, (a1, a2) in enumerate(zip(arr1.T, arr2.T)):
            jpd[gids[i]] = np.histogram2d(a1, a2,
                                          bins=(bins1, bins2),
                                          density=True)[0].astype(np.float32)

        return jpd

    def _get_slices(self, dset1, dset2, sites=None, chunks_per_slice=5):
        """
        Get slices to extract, ensure the shapes of dset1 and 2 match.

        Parameters
        ----------
        dset1 : str
            Dataset 1 to generate joint probability distribution for
        dset2 : str
            Dataset 2 to generate joint probabilty distribution for
        sites : list | slice, optional
            Subset of sites to extract, by default None or all sites
        chunks_per_slice : int, optional
            Number of chunks to extract in each slice, by default 5

        Returns
        -------
        slices : list
            List of slices to extract
        """
        with self.res_cls(self.res_h5) as f:
            shape, _, chunks = f.get_dset_properties(dset1)
            shape2, _, _ = f.get_dset_properties(dset2)

        if shape != shape2:
            msg = ("The shape of {}: {}, does not match the shape of {}: {}!"
                   .format(dset1, shape, dset2, shape2))
            logger.error(msg)
            raise RuntimeError(msg)

        slices = slice_sites(shape, chunks, sites=sites,
                             chunks_per_slice=chunks_per_slice)

        return slices

    def compute(self, dset1, dset2, bins1, bins2, sites=None, max_workers=None,
                chunks_per_worker=5):
        """
        Compute joint probability distribution between given datasets using
        given bins for all sites.

        Parameters
        ----------
        dset1 : str
            Dataset 1 to generate joint probability distribution for
        dset2 : str
            Dataset 2 to generate joint probabilty distribution for
        bins1 : tuple
            (start, stop, step) for dataset 1 bins. The stop value is
            inclusive, so (0, 6, 2) would yield three bins with edges (0, 2, 4,
            6). If the stop value is not perfectly divisible by the step, the
            last bin will overshoot the stop value.
        bins2 : tuple
            (start, stop, step) for dataset 2 bins. The stop value is
            inclusive, so (0, 6, 2) would yield three bins with edges (0, 2, 4,
            6). If the stop value is not perfectly divisible by the step, the
            last bin will overshoot the stop value.
        sites : list | slice, optional
            Subset of sites to extract, by default None or all sites
        max_workers : None | int, optional
            Number of workers to use, if 1 run in serial, if None use all
            available cores, by default None
        chunks_per_worker : int, optional
            Number of chunks to extract on each worker, by default 5

        Returns
        -------
        jpd: pandas.DataFrame
            DataFrame of joint probability distribution between given datasets
            with given bins
        """
        if max_workers is None:
            max_workers = os.cpu_count()

        slices = self._get_slices(dset1, dset2, sites,
                                  chunks_per_slice=chunks_per_worker)
        if len(slices) == 1:
            max_workers = 1

        jpd = {}
        if max_workers > 1:
            msg = ('Computing the joint probability distribution between {} '
                   'and {} in parallel using {} workers'
                   .format(dset1, dset2, max_workers))
            logger.info(msg)

            loggers = [__name__, 'rex']
            with SpawnProcessPool(max_workers=max_workers,
                                  loggers=loggers) as exe:
                futures = []
                for sites_slice in slices:
                    future = exe.submit(self.compute_joint_pd,
                                        self.res_h5, dset1, dset2,
                                        bins1, bins2,
                                        res_cls=self.res_cls,
                                        hsds=self._hsds,
                                        sites_slice=sites_slice)
                    futures.append(future)

                for i, future in enumerate(as_completed(futures)):
                    jpd.update(future.result())
                    logger.debug('Completed {} out of {} workers'
                                 .format((i + 1), len(futures)))

        else:
            msg = ('Computing the joint probability distribution between {} '
                   'and {} in serial.'
                   .format(dset1, dset2))
            logger.info(msg)
            for i, sites_slice in enumerate(slices):
                jpd.update(self.compute_joint_pd(
                    self.res_h5, dset1, dset2,
                    bins1, bins2,
                    res_cls=self.res_cls,
                    hsds=self._hsds,
                    sites_slice=sites_slice))
                logger.debug('Completed {} out of {} sets of sites'
                             .format((i + 1), len(slices)))

        gc.collect()
        log_mem(logger)
        bins1 = self._make_bins(*bins1)
        bins2 = self._make_bins(*bins2)
        index = np.meshgrid(bins1[:-1], bins2[:-1], indexing='ij')
        index = np.array(index).T.reshape(-1, 2).astype(np.int16)
        index = pd.MultiIndex.from_arrays(index.T, names=(dset1, dset2))
        jpd = pd.DataFrame({k: v.flatten(order='F') for k, v
                            in jpd.items()}, index=index).sort_index(axis=1)

        return jpd

    def save(self, jpd, out_fpath):
        """
        Save joint probability distribution to disk

        Parameters
        ----------
        jpd : pandas.DataFrame
            Table of joint probability distribution densities to save
        out_fpath : str
            .csv, or .h5 file path to save joint probability
            distribution to
        """
        with self.res_cls(self.res_h5) as f:
            meta = f['meta', jpd.columns.values]

        logger.info('Writing joint probability distribution to {}'
                    .format(out_fpath))
        if out_fpath.endswith('.csv'):
            jpd.to_csv(out_fpath)
            meta_fpath = out_fpath.split('.')[0] + '_meta.csv'
            if os.path.exists(meta_fpath):
                msg = ("Site meta data already exists at {}!")
                logger.warning(msg)
                warn(msg)
            else:
                logger.debug('Writing site meta data to {}'
                             .format(meta_fpath))
                meta.to_csv(meta_fpath, index=False)
        elif out_fpath.endswith('.h5'):
            with h5py.File(out_fpath, mode='w') as f:
                f.attrs['rex version'] = __version__
                for i, n in enumerate(jpd.index.names):
                    logger.info('')
                    data = np.array(jpd.index.get_level_values(i))
                    dset = '{}-bins'.format(n)
                    logger.debug('Writing {}'.format(dset))
                    f.create_dataset(dset, shape=data.shape, dtype=data.dtype,
                                     data=data)

                logger.debug('Writing joint probability density values to jpd')
                data = jpd.values
                f.create_dataset('jpd', shape=data.shape, dtype=data.dtype,
                                 data=data)

                logger.debug('Writing site meta data to meta')
                meta = to_records_array(meta)
                f.create_dataset('meta', shape=meta.shape, dtype=meta.dtype,
                                 data=meta)
        else:
            msg = ("Cannot save joint probability distribution, expecting "
                   ".csv or .h5 path, but got: {}".format(out_fpath))
            logger.error(msg)
            raise OSError(msg)

    @staticmethod
    def plot_joint_pd(jpd, site=None, **kwargs):
        """
        Plot the mean joint probability distribution accross all sites
        (site=None), or the distribution for the single given site

        Parameters
        ----------
        jpd: pandas.DataFrame
            DataFrame of joint probability distribution between given datasets
            with given bins
        site : int, optional
            Site to plot distribution for, if None plot mean distribution
            across all sites, by default None
        """
        x, y = jpd.index.names
        if site is not None:
            msg = ("Can only plot the joint probabilty distribution for a "
                   "single site or the mean probability distribution accross "
                   "all sites (site=None), you provided: {}".format(site))
            assert isinstance(site), msg
            plt = jpd.loc[:, [site]].reset_index()
        else:
            site = 'mean_jpd'
            plt = jpd.mean(axis=1)
            plt.name = site
            plt = plt.reset_index()

        plt.plot.scatter(x=x, y=y, c=site, **kwargs)

    @classmethod
    def run(cls, res_h5, dset1, dset2, bins1, bins2,
            sites=None, res_cls=Resource, hsds=False,
            max_workers=None, chunks_per_worker=5, out_fpath=None):
        """
        Compute joint probability distribution between given datasets using
        given bins

        Parameters
        ----------
        res_h5 : str
            Path to resource h5 file(s)
        dset1 : str
            Dataset 1 to generate joint probability distribution for
        dset2 : str
            Dataset 2 to generate joint probabilty distribution for
        bins1 : tuple
            (start, stop, step) for dataset 1 bins. The stop value is
            inclusive, so (0, 6, 2) would yield three bins with edges (0, 2, 4,
            6). If the stop value is not perfectly divisible by the step, the
            last bin will overshoot the stop value.
        bins2 : tuple
            (start, stop, step) for dataset 2 bins. The stop value is
            inclusive, so (0, 6, 2) would yield three bins with edges (0, 2, 4,
            6). If the stop value is not perfectly divisible by the step, the
            last bin will overshoot the stop value.
        sites : list | slice, optional
            Subset of sites to extract, by default None or all sites
        res_cls : Class, optional
            Resource class to use to access res_h5, by default Resource
        hsds : bool, optional
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS, by default False
        max_workers : None | int, optional
            Number of workers to use, if 1 run in serial, if None use all
            available cores, by default None
        chunks_per_worker : int, optional
            Number of chunks to extract on each worker, by default 5
        out_fpath : str, optional
            .csv, or .h5 file path to save joint probability
            distribution to

        Returns
        -------
        out : pandas.DataFrame
            DataFrame of joint probability distribution between given datasets
            with given bins
        """
        logger.info('Computing joint probability distribution between {} and '
                    '{} in {}'
                    .format(dset1, dset2, res_h5))
        logger.debug('Computing joint probability distribution using:'
                     '\n-{} bins: {}'
                     '\n-{} bins: {}'
                     '\n-max workers: {}'
                     '\n-chunks per worker: {}'
                     .format(dset1, bins1, dset2, bins2, max_workers,
                             chunks_per_worker))
        jpd = cls(res_h5, res_cls=res_cls, hsds=hsds)
        out = jpd.compute(dset1, dset2, bins1, bins2,
                          sites=sites,
                          max_workers=max_workers,
                          chunks_per_worker=chunks_per_worker)
        if out_fpath is not None:
            jpd.save(out, out_fpath)

        return out

    @classmethod
    def wind_rose(cls, wind_h5, hub_height, wspd_bins=(0, 30, 1),
                  wdir_bins=(0, 360, 5), sites=None, res_cls=WindResource,
                  hsds=False, max_workers=None, chunks_per_worker=5,
                  out_fpath=None):
        """
        Compute wind rose at given hub height

        Parameters
        ----------
        wind_h5 : str
            Path to resource h5 file(s)
        hub_height : str | int
            Hub-height to compute wind rose at
        wspd_bins : tuple
            (start, stop, step) for wind speed bins
        wdir_bins : tuple
            (start, stop, step) for wind direction bins
        sites : list | slice, optional
            Subset of sites to extract, by default None or all sites
        res_cls : Class, optional
            Resource class to use to access wind_h5, by default Resource
        hsds : bool, optional
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS, by default False
        max_workers : None | int, optional
            Number of workers to use, if 1 run in serial, if None use all
            available cores, by default None
        chunks_per_worker : int, optional
            Number of chunks to extract on each worker, by default 5
        out_fpath : str, optional
            .csv, or .h5 file path to save wind rose to

        Returns
        -------
        wind_rose : pandas.DataFrame
            DataFrame of wind rose frequencies at desired hub-height
        """
        logger.info('Computing wind rose for {}m wind in {}'
                    .format(hub_height, wind_h5))
        logger.debug('Computing wind rose using:'
                     '\n-wind speed bins: {}'
                     '\n-wind direction bins: {}'
                     '\n-max workers: {}'
                     '\n-chunks per worker: {}'
                     .format(wspd_bins, wdir_bins, max_workers,
                             chunks_per_worker))
        wind_rose = cls(wind_h5, res_cls=res_cls, hsds=hsds)
        wspd_dset = 'windspeed_{}m'.format(hub_height)
        wdir_dset = 'winddirection_{}m'.format(hub_height)
        out = wind_rose.compute(wspd_dset, wdir_dset, wspd_bins, wdir_bins,
                                sites=sites,
                                max_workers=max_workers,
                                chunks_per_worker=chunks_per_worker)
        if out_fpath is not None:
            wind_rose.save(out, out_fpath)

        return out



"""
Module to rechunk existing .h5 files
"""
from abc import ABC
from math import ceil
import numpy as np


class BaseChunkSize(ABC):
    """
    Base class to compute chunk size
    """
    def __init__(self):
        """
        Parameters
        ----------
        chunk_size : int, optional
            Chunk size in MB, by default 2
        """
        self._chunks = None

    @property
    def chunks(self):
        """
        Dataset chunk size along all axis

        Returns
        -------
        tuple
        """
        return self._chunks


class TimeseriesChunkSize(BaseChunkSize):
    """
    Compute Timeseries chunks based on dtype, and weeks_per_chunk
    """
    def __init__(self, shape, dtype, chunk_size=2, weeks_per_chunk=None):
        """
        Parameters
        ----------
        shape : tuple
            Array shape
        dtype : str | np.dtype
            Array data type
        chunk_size : int, optional
            Chunk size in MB, by default 2
        weeks_per_chunk : int, optional
            Number of weeks per time chunk, if None scale weeks based on 8
            weeks for hourly data, by default None
        """
        self._chunks = self.compute_dtype_chunks(
            shape, dtype,
            chunk_size=chunk_size,
            weeks_per_chunk=weeks_per_chunk)

    @staticmethod
    def _compute_time_chunk_size(t_len, weeks_per_chunk=8):
        """
        Compute the size of the temporal chunks. Chunks are based on

        Parameters
        ----------
        t_len : int
            Length of temporal axis
        weeks_per_chunk : int, optional
            Number of weeks per time chunk, if None scale weeks based on 8
            weeks for hourly data, by default None

        Returns
        -------
        time_chunks : int
            Size of chunks along temporal axis
        """
        # Freq is the number of time-steps per hour
        freq = t_len / 8760

        if weeks_per_chunk is None:
            # Use default of 8 weeks for hourly chunks
            weeks_per_chunk = ceil(8 / freq)

        # Compute time chunks
        time_chunk = ceil(weeks_per_chunk * 7 * 24 * freq)

        if time_chunk >= t_len:
            time_chunk = ceil(t_len / 4)

        return time_chunk

    @staticmethod
    def _compute_site_chunk_size(time_chunk, dtype, chunk_size=2):
        """
        Compute spatial chunk size given time chunk size

        Parameters
        ----------
        time_chunk : int
            Lenght of chunk along the temporal axis
        dtype : str | np.dtype
            Array data type (dtype)
        chunk_size : int, optional
            Chunk size in MB, by default 2

        Returns
        -------
        site_chunk : int
            Spatial chunk size or size of chunk along non-temporal axes
        """
        pixel_size = np.dtype(dtype).itemsize * 10**-6
        site_chunk = ceil(chunk_size / (time_chunk * pixel_size))

        return site_chunk

    @classmethod
    def compute_dtype_chunks(cls, shape, dtype, chunk_size=2,
                             weeks_per_chunk=None):
        """
        Compute chunks based on array shape and dtype

        Parameters
        ----------
        shape : tuple
            Array shape
        dtype : str | np.dtype
            Array data type
        chunk_size : int, optional
            Chunk size in MB, by default 2
        weeks_per_chunk : int, optional
            Number of weeks per time chunk, if None scale weeks based on 8
            weeks for hourly data, by default None

        Returns
        -------
        chunks : tuple
            Dataset chunk size along all axis
        """
        time_chunk = cls._compute_time_chunk_size(
            shape[0],
            weeks_per_chunk=weeks_per_chunk)
        site_chunk = cls._compute_site_chunk_size(time_chunk, dtype,
                                                  chunk_size=chunk_size)

        sites = np.sum(shape[1:])
        if site_chunk >= sites:
            site_chunk = shape[1:]
        elif len(shape) > 2:
            weights = [i // sites for i in shape[1:]]
            site_chunk = (ceil(w * site_chunk) for w in weights)
        else:
            site_chunk = (site_chunk, )

        chunks = (time_chunk, ) + site_chunk

        return chunks

    @classmethod
    def compute(cls, shape, dtype, chunk_size=2, weeks_per_chunk=None):
        """
        Compute chunks based on array shape and dtype

        Parameters
        ----------
        shape : tuple
            Array shape
        dtype : str | np.dtype
            Array data type
        chunk_size : int, optional
            Chunk size in MB, by default 2
        weeks_per_chunk : int, optional
            Number of weeks per time chunk, if None scale weeks based on 8
            weeks for hourly data, by default None

        Returns
        -------
        chunks : tuple
            Dataset chunk size along all axis
        """
        chunks = cls(shape, dtype, chunk_size=chunk_size,
                     weeks_per_chunk=weeks_per_chunk)

        return chunks.chunks


class ArrayChunkSize(BaseChunkSize):
    """
    Compute chunks based on array size, array will only be chunked along the
    0 axis
    """
    def __init__(self, arr, chunk_size=2):
        """
        Parameters
        ----------
        arr : ndarray
            Dataset array
        chunk_size : int, optional
            Chunk size in MB, by default 2
        """
        self._chunks = self.compute_arr_chunks(arr, chunk_size=chunk_size)

    @staticmethod
    def _get_arr_size(arr):
        """
        Get the size of the given array in MB

        Parameters
        ----------
        data : ndarray
            Data array to chunk

        Returns
        -------
        float
            Data array size in MB
        """

        return arr.size * arr.dtype.itemsize * 10**-6

    @classmethod
    def compute_arr_chunks(cls, arr, chunk_size=2):
        """
        Compute chunks based on array size, array will only be chunked along
        the 0 axis

        Parameters
        ----------
        arr : ndarray
            Dataset array
        chunk_size : int, optional
            Chunk size in MB, by default 2

        Returns
        -------
        chunks : tuple
            Dataset chunk size along all axis
        """
        arr_size = cls._get_arr_size(arr)
        if arr_size >= 2:
            arr_len = len(arr)
            chunks = ceil(arr_len // (arr_size / chunk_size))
            chunks = (chunks, ) + arr.shape[1:]
        else:
            chunks = None

        return chunks

    @classmethod
    def compute(cls, arr, chunk_size=2):
        """
        Compute chunks based on array size, array will only be chunked along
        the 0 axis

        Parameters
        ----------
        arr : ndarray
            Dataset array
        chunk_size : int, optional
            Chunk size in MB, by default 2

        Returns
        -------
        chunks : tuple
            Dataset chunk size along all axis
        """
        chunks = cls(arr, chunk_size=chunk_size)

        return chunks.chunks



# -*- coding: utf-8 -*-
"""
Collection of helpful functions
"""
import datetime
import inspect
import json
import yaml
import os
from fnmatch import fnmatch
import numpy as np
import pandas as pd
from pandas.api.types import CategoricalDtype
import re
from scipy.spatial import cKDTree
import time
from warnings import warn
from packaging import version


from rex.utilities.exceptions import (FileInputError, JSONError, RetryError,
                                      RetryWarning)


def safe_json_load(fpath):
    """Perform a json file load with better exception handling.

    Parameters
    ----------
    fpath : str
        Filepath to .json file.

    Returns
    -------
    dict
        Loaded json dictionary.

    Examples
    --------
    >>> json_path = "./path_to_json.json"
    >>> safe_json_load(json_path)
    {key1: value1,
     key2: value2}
    """

    validate_filepath(fpath, file_extension='.json', exception_type=JSONError)
    return _read_data_file(fpath, json.load, exception_type=JSONError)


def safe_yaml_load(fpath):
    """Perform a yaml file load with better exception handling.

    Parameters
    ----------
    fpath : str
        Filepath to .yaml (or .yml) file.

    Returns
    -------
    dict
        Loaded yaml dictionary.

    Examples
    --------
    >>> yaml_path = "./path_to_yaml.yaml"
    >>> safe_yaml_load(yaml_path)
    {key1: value1,
     key2: value2}
    """

    validate_filepath(fpath, file_extension=('.yml', '.yaml'),
                      exception_type=yaml.YAMLError)
    return _read_data_file(fpath, yaml.safe_load,
                           exception_type=yaml.YAMLError)


def validate_filepath(fpath, file_extension, exception_type):
    """Validate an input filepath.

    The input is verified to be a string with a valid ending, and
    it is validated that the file exists on disk. If any of these conditions
    are not met, an exception is raised.


    Parameters
    ----------
    fpath : str
        Filepath to validate.
    file_extension : str or iterable of str
        A single file extension or an iterable of acceptable
        file extensions for the input path.
    exception_type : `Exception`
        A class indicating the type of exception to raise if
        file extension is incorrect.

    Raises
    ------
    TypeError
        If the input `fpath` is not a string.
    exception_type
        If the input `fpath` does not end in a valid extension.
    FileNotFoundError
        If the input `fpath` does not exist on disk.
    """

    if not isinstance(fpath, str):
        raise TypeError('Filepath must be str to load: {}'.format(fpath))

    if not fpath.endswith(file_extension):
        raise exception_type('Filepath must end in {!r} to load: {}'
                             .format(file_extension, fpath))

    if not os.path.isfile(fpath):
        raise FileNotFoundError('Could not find file to load: {}'
                                .format(fpath))


def _read_data_file(fpath, load_method, exception_type):
    """Load the data in the file using a given load method.

    This function performs additional exception handling during the
    data loading process.

    Parameters
    ----------
    fpath : str
        Filepath containing data to load.
    load_method : callable
        Function that can be called on a stream to load the
        data it contains.
    exception_type : `Exception`
        A class indicating the type of exception to raise if
        data cannot be read.

    Returns
    -------
    data : dict
        Dictionary representation of the data in the file.

    Raises
    ------
    exception_type
        If there was an error loading the data.
    """

    try:
        with open(fpath, 'r') as f:
            data = load_method(f)
    except exception_type as e:
        msg = 'Error:\n{}\nCannot read file: "{}"'.format(e, fpath)
        raise exception_type(msg) from e

    return data


def jsonify_dict(di):
    """Jsonify a dictionary into a string with handling for int/float keys.

    Parameters
    ----------
    di : dict
        Dictionary to be jsonified.

    Returns
    -------
    sdi : str
        Jsonified dictionary. Int/float keys will be represented as strings
        because json objects outside of python cannot have int/float keys.
    """

    for k in list(di.keys()):
        try:
            float(k)
        except ValueError:
            pass
        else:
            di[str(k)] = di.pop(k)

    try:
        sdi = json.dumps(di)
    except TypeError as e:
        msg = ('Could not json serialize {}, received error: {}'
               .format(di, e))
        raise TypeError(msg) from e

    return sdi


def dict_str_load(dict_str):
    """
    Load jsonified string entries into dictionaries using JSON

    Parameters
    ----------
    dict_str : str
        JSON style string provided to CLI or in config

    Returns
    -------
    out_dict : dict
        Dictionary loaded by JSON

    Examples
    --------
    >>> json_str = "{bool_key: 'True', value_key: 'None'}"
    >>> dict_str_load(json_str)
    {bool_key: True,
     value_key: None}
    """
    dict_str = dict_str.replace('\'', '\"')
    dict_str = dict_str.replace('None', 'null')
    dict_str = dict_str.replace('True', 'true')
    dict_str = dict_str.replace('False', 'false')
    out_dict = json.loads(dict_str)

    return out_dict


def parse_year(inp, option='raise'):
    """
    Attempt to parse a year out of a string.

    Parameters
    ----------
    inp : str
        String from which year is to be parsed
    option : str
        Return option:
         - "bool" will return True if year is found, else False.
         - Return year int / raise a RuntimeError otherwise

    Returns
    -------
    out : int | bool
        Year int parsed from inp,
        or boolean T/F (if found and option is bool).

    Examples
    --------
    >>> year_str = "NSRDB_2018.h5"
    >>> parse_year(year_str)
    2018

    >>> year_str = "NSRDB_2018.h5"
    >>> parse_year(year_str, option='bool')
    True

    >>> year_str = "NSRDB_TMY.h5"
    >>> parse_year(year_str)
    RuntimeError: Cannot parse year from NSRDB_TMY.h5

    >>> year_str = "NSRDB_TMY.h5"
    >>> parse_year(year_str, option='bool')
    False
    """
    # char leading year cannot be 0-9
    # char trailing year can be end of str or not 0-9
    regex = r".*[^0-9]([1-2][0-9]{3})($|[^0-9])"

    match = re.match(regex, inp)

    if match:
        out = int(match.group(1))

        if 'bool' in option:
            out = True

    else:
        if 'bool' in option:
            out = False
        else:
            raise RuntimeError('Cannot parse year from {}'.format(inp))

    return out


def mean_irrad(arr):
    """Calc the annual irradiance at a site given an irradiance timeseries.

    Parameters
    ----------
    arr : np.ndarray | pd.Series
        Annual irradiance array in W/m2. Row dimension is time.

    Returns
    -------
    mean : float | np.ndarray
        Mean irradiance values in kWh/m2/day. Float if the input array is
        1D, 1darray if the input array is 2D (multi-site).
    """

    mean = arr.mean(axis=0) / 1000 * 24
    return mean


def check_res_file(res_file):
    """
    Check resource to see if the given path
    - It belongs to a multi-file handler
    - Is on local disk
    - Is a hsds path

    Parameters
    ----------
    res_file : str
        Filepath to single resource file, unix style multi-file path like
        /h5_dir/prefix*suffix.h5, or an hsds filepath (filename of hsds
        path can also contain wildcards *)

    Returns
    -------
    multi_h5_res : bool
        Boolean flag to use a MultiFileResource handler
    hsds : bool
        Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
        behind HSDS
    """
    multi_h5_res = False
    hsds = False

    if os.path.isfile(res_file):
        pass

    elif '*' in res_file:
        multi_h5_res = True

    elif os.path.isdir(res_file):
        msg = ('Cannot parse directory, need to add wildcard * suffix: {}'
               .format(res_file))
        raise FileInputError(msg)

    else:
        try:
            import h5pyd
            hsds_dir = os.path.dirname(res_file)
            with h5pyd.Folder(hsds_dir + '/') as f:
                hsds = True
                fps = [f'{hsds_dir}/{fn}' for fn in f
                       if fnmatch(f'{hsds_dir}/{fn}', res_file)]
                if not any(fps):
                    msg = ('{} is not a valid HSDS file path!'
                           .format(res_file))
                    raise FileNotFoundError(msg)
                elif len(fps) > 1:
                    multi_h5_res = True

        except Exception as ex:
            msg = ("{} is not a valid file path, and HSDS "
                   "cannot be check for a file at this path:{}!"
                   .format(res_file, ex))
            raise FileNotFoundError(msg) from ex

    return multi_h5_res, hsds


def parse_date_int(s):
    """Parse data parameters from an integer or string of format YYYYMMDD

    Parmeters
    ---------
    s : str | int
        Date string or integer of format YYYYMMDD

    Returns
    -------
    y : int
        Year integer parsed from input.
    m : int
        Month integer parsed from input.
    d : int
        Day integer parsed from input.
    """

    try:
        s = str(int(s))
    except ValueError as ex:
        e = ('Could not convert date string to int: "{}"'
             .format(s))
        raise ValueError(e) from ex

    assert len(s) == 8, 'Bad date string, should be YYYYMMDD: {}'.format(s)

    y = int(s[0:4])
    m = int(s[4:6])
    d = int(s[6:8])

    assert y > 1970, 'Bad date string, year < 1970: {}'.format(s)
    assert m < 13, 'Bad date string, month > 12: {}'.format(s)
    assert d < 32, 'Bad date string, day > 31: {}'.format(s)

    return y, m, d


def str_to_date(s):
    """Convert a date string of format YYYYMMDD to date object.

    Parameters
    ----------
    s : str
        Date string of format YYYYMMDD

    Returns
    -------
    d : datetime.date
        Date object.
    """
    d = datetime.date(*parse_date_int(s))
    return d


def str_to_datetime(s):
    """Convert a date string of format YYYYMMDD to datetime object.

    Parameters
    ----------
    s : str
        Date string of format YYYYMMDD

    Returns
    -------
    d : datetime.datetime
        Datetime object.
    """
    d = datetime.datetime(*parse_date_int(s))
    return d


def parse_table(table):
    """
    Load pandas DataFrame from .csv or .json file or dictionary

    Parameters
    ----------
    trans_table : str | pandas.DataFrame | dict
        Path to .csv or .json or dictionary containing table to parse

    Returns
    -------
    table : pandas.DataFrame
        DataFrame table
    """
    if isinstance(table, str):
        if table.endswith('.csv'):
            table = pd.read_csv(table)
            if 'Unnamed: 0' in table:
                table = table.drop(columns='Unnamed: 0')

        elif table.endswith('.json'):
            table = pd.read_json(table)
        else:
            raise ValueError('Cannot parse {}, expecting a .csv or .json file'
                             .format(table))
    elif isinstance(table, dict):
        table = pd.DataFrame(table)
    elif not isinstance(table, pd.DataFrame):
        raise ValueError('Cannot parse table from type {}, expecting a .csv, '
                         '.json, dictionary, or pandas.DataFrame'
                         .format(type(table)))

    return table


def get_class_properties(cls):
    """
    Get all class properties
    Used to check against config keys

    Returns
    -------
    properties : list
        List of class properties, each of which should represent a valid
        config key/entry
    """
    properties = [attr for attr, attr_obj
                  in inspect.getmembers(cls)
                  if isinstance(attr_obj, property)]

    return properties


def timestamp_format_to_redex(time_format):
    """
    convert time stamp format to redex

    Parameters
    ----------
    time_format : str
        datetime timestamp format

    Returns
    -------
    redex : str
        redex format for timestamp
    """

    time_keys = {'%Y': r'\d{4}',
                 '%m': r'\d{2}',
                 '%d': r'\d{2}',
                 '%H': r'\d{2}',
                 '%M': r'\d{2}',
                 '%S': r'\d{2}'}

    redex = time_format
    for key, item in time_keys.items():
        if key in redex:
            redex = redex.replace(key, item)

    return redex


def parse_timestamp(path, time_format='%Y-%m-%d_%H:%M:%S'):
    """
    extract timestamp with given format from given path

    Parameters
    ----------
    path : str
        file path
    time_format : str, optional
        datetime timestamp format, by default '%Y-%m-%d_%H:%M:%S'

    Returns
    -------
    str
        Portion of path that matches given format
    """
    pattern = timestamp_format_to_redex(time_format)
    pattern = re.compile(pattern)
    matcher = pattern.search(path)

    if matcher is None:
        raise RuntimeError("Could not find timestamp with format {} in {}!"
                           .format(time_format, path))

    return matcher.group()


def filename_timestamp(file_name, time_format='%Y-%m-%d_%H:%M:%S'):
    """
    extract timestamp from file name

    Parameters
    ----------
    file_name : str
        file name or file path
    time_format : str, optional
        datetime timestamp format, by default '%Y-%m-%d_%H:%M:%S'

    Returns
    -------
    str
        Portion of file_name that matches given format
    """
    timestamp = parse_timestamp(os.path.basename(file_name),
                                time_format=time_format)

    return timestamp


class Retry:
    """
    Retry Decorator to run a function multiple times
    """
    def __init__(self, tries=3, n_sec=1):
        """
        Parameters
        ----------
        tries : int, optional
            Number if times to retry function, by default 2
        n_sec : int, optional
            Number of seconds to wait between tries, by default 1
        """
        self._tries = tries
        self._wait = n_sec

    def __call__(self, func, *args, **kwargs):
        """
        Decorator call

        Parameters
        ----------
        func : obj
            Function to retry on Exception
        args : tuple
            Function arguments
        kwargs : dict
            Function kwargs
        """
        def new_func(*args, **kwargs):
            i = 0
            error = None
            while i <= self._tries:
                try:
                    new_func = func(*args, **kwargs)
                    break
                except RetryError as ex:
                    msg = ('{} failed to run {} times:\n{}'
                           .format(func.__name__, i, ex))
                    raise RuntimeError(msg) from ex
                except Exception as ex:
                    error = ex
                    warn('Attempt {} failed:\n{}'.format(i, error),
                         RetryWarning)
                    time.sleep(self._wait)
                finally:
                    i += 1

            if i > self._tries:
                raise RetryError('Failed to run {}:\n{}'
                                 .format(func.__name__, error))

            return new_func

        return new_func


def check_eval_str(s):
    """Check an eval() string for questionable code.

    Parameters
    ----------
    s : str
        String to be sent to eval(). This is most likely a math equation to be
        evaluated. It will be checked for questionable code like imports and
        dunder statements.
    """
    bad_strings = ('import', 'os.', 'sys.', '.__', '__.')
    for bad_s in bad_strings:
        if bad_s in s:
            raise ValueError('Will not eval() string which contains "{}": {}'
                             .format(bad_s, s))


def check_tz(time_index):
    """
    Check datetime index for timezone, if None set to UTC

    Parameters
    ----------
    time_index : pandas.DatatimeIndex
        DatetimeIndex to check timezone for

    Returns
    -------
    time_index : pandas.DatatimeIndex
        Updated DatetimeIndex with timezone set
    """
    if not time_index.tz:
        time_index = time_index.tz_localize('utc')

    return time_index


def get_lat_lon_cols(df):
    """
    Get columns that contain (latitude, longitude) coordinates

    Parameters
    ----------
    df : pandas.DataFrame
        DataFrame to extract coordinates (lat, lon) from

    Returns
    -------
    lat_lon_cols : list
        Column names in df that correspond to the latitude and longitude
        coordinates. There must be a single unique set of latitude and
        longitude columns.
    """
    lat_lon_cols = ['latitude', 'longitude']
    lat = False
    lon = False
    for c in df.columns:
        if c.lower() in ['lat', 'latitude']:
            if lat:
                msg = ("Multiple possible latitude columns were found: "
                       "({}, {})!".format(lat_lon_cols[0], c))
                raise RuntimeError(msg)

            lat_lon_cols[0] = c
            lat = True
        elif c.lower() in ['lon', 'long', 'longitude']:
            if lon:
                msg = ("Multiple possible longitude columns were found: "
                       "({}, {})!".format(lat_lon_cols[1], c))
                raise RuntimeError(msg)

            lat_lon_cols[1] = c
            lon = True

    if not lat or not lon:
        msg = ("A valid pair of latitude and longitude columns could not be "
               "found in: {}!".format(df.columns))
        raise RuntimeError(msg)

    return lat_lon_cols


def roll_timeseries(arr, timezones):
    """
    Roll timeseries from UTC to local time. Automatically compute time-shift
    from UTC offset (timezone) and time-series length.

    Parameters
    ----------
    arr : ndarray
        Input timeseries array of form (time, sites)
    timezones : ndarray | list
        Vector of timezone shifts from UTC to local time

    Returns
    -------
    local_arr : ndarray
        Array shifted to local time
    """
    if arr.shape[1] != len(timezones):
        msg = ('Number of timezone shifts ({}) does not match number of '
               'sites ({})'.format(len(timezones), arr.shape[1]))
        raise ValueError(msg)

    time_step = arr.shape[0] // 8760

    local_arr = np.zeros(arr.shape, dtype=arr.dtype)
    for tz in set(timezones):
        mask = timezones == tz
        local_arr[:, mask] = np.roll(arr[:, mask], int(tz * time_step), axis=0)

    return local_arr


def get_chunk_ranges(ds_dim, chunk_size):
    """
    Create list of chunk slices [(s_i, e_i), ...]

    Parameters
    ----------
    ds_len : int
        Length of dataset axis to chunk
    chunk_size : int
        Size of chunks

    Returns
    -------
    chunks : list
        List of chunk start and end positions
        [(s_i, e_i), (s_i+1, e_i+1), ...]
    """
    chunks = list(range(0, ds_dim, chunk_size))
    if chunks[-1] < ds_dim:
        chunks.append(ds_dim)
    else:
        chunks[-1] = ds_dim

    chunks = list(zip(chunks[:-1], chunks[1:]))

    return chunks


def split_sites_slice(sites_slice, n_sites, slice_size):
    """
    Break up sites_slice into slices of size slice_size

    Parameters
    ----------
    sites_slice : slice
        Sites to extract as a slice object to extract
    n_sites : int
        Total number of sites to extract
    slice_size : int
        Number of sites in each slice to extract either on each worker,
        or in series

    Returns
    -------
    slices : list
        List of slices to extract
    """
    stop = sites_slice.stop
    if stop is None:
        stop = n_sites

    if slice_size >= n_sites:
        msg = ('The slice_size {} is >= the number of sites to be '
               'extracted {}! A single slice will be extracted.'
               .format(slice_size, n_sites))
        warn(msg)

        slices = [slice(sites_slice.start, stop, sites_slice.step)]
    else:
        step = sites_slice.step
        if step is not None:
            slice_size *= step

        # Create slices of size slice_size
        slices = [slice(s, e, step) for s, e
                  in get_chunk_ranges(stop, slice_size)]

    return slices


def split_sites_list(sites, slice_size):
    """
    Split sites into sub-lists of ~ size slice_size

    Parameters
    ----------
    sites : list
        Sites to extract as a list or numpy object to extract
    slice_size : int
        Number of sites in each slice to extract either on each worker,
        or in series

    Returns
    -------
    slices : list
        List of slices to extract
    """
    if slice_size >= len(sites):
        msg = ('The slice_size {} is >= the number of sites to be '
               'extracted {}! A single slice will be extracted.'
               .format(slice_size, len(sites)))
        warn(msg)
        slices = [sites]
    else:
        slices = np.array_split(sites, len(sites) // slice_size)

    return slices


def slice_sites(shape, chunks, sites=None, chunks_per_slice=5):
    """
    Slice sites into given number of sub-sets with given number of chunks per
    sub-set

    Parameters
    ----------
    shape : tuple
        Shape of dataset array that data is being extracted from
    chunks : tuple
        Chunk size of dataset array in .h5 file from which dataset is being
        extracted
    sites : list | slice, optional
        Subset of sites to extract, by default None or all sites
    chunks_per_slice : int, optional
        Number of chunks to extract in each slice, by default 5

    Returns
    -------
    slices : list
        List of slices to extract
    """
    if chunks is not None:
        slice_size = chunks[1] * chunks_per_slice
    else:
        slice_size = chunks_per_slice * 100

    if sites is None:
        sites = slice(None)

    if isinstance(sites, slice):
        slices = split_sites_slice(sites, shape[1], slice_size)
    elif isinstance(sites, (list, tuple, np.ndarray)):
        slices = split_sites_list(sites, slice_size)
    else:
        msg = ('sites must be of type "None", "slice", "list", "tuple", '
               'or "np.ndarray", but {} was provided'.format(type(sites)))
        raise TypeError(msg)

    return slices


def res_dist_threshold(lat_lons, tree=None, margin=1.05):
    """
    Distance threshold for nearest neighbor searches performed on resource
    points. Calculated as half of the diagonal between closest resource points,
    with desired extra margin

    Parameters
    ----------
    lat_lons : ndarray
        n x 2 array of resource points coordinates (lat, lon)
    tree : cKDTree, optional
        Pre-build cKDTree of resource lat, lon coordintes. If None, build the
        cKDTree from scratch, by default None
    margin : float, optional
        Extra margin to multiply times the computed max distance between
        neighboring resource points, by default 1.05

    Returns
    -------
    float
        Distance threshold for nearest neighbor searches performed on resource
        points. Calculated as half of the diagonal between closest resource
        points, with desired extra margin
    """
    if tree is None:
        # pylint: disable=not-callable
        tree = cKDTree(lat_lons)

    dists = tree.query(lat_lons, k=2)[0][:, 1]
    dists = dists[(dists != 0)]

    return margin * (2 ** 0.5) * (dists.max() / 2)


def get_dtype(col):
    """
    Get column dtype for converstion to records array

    Parameters
    ----------
    col : pandas.Series
        Column from pandas DataFrame

    Returns
    -------
    out : str
        String representation of converted dtype for column:
        -  float = float32
        -  int = int16 or int32 depending on data range
        -  object/str = U* max length of strings in col
    """
    dtype = col.dtype

    if isinstance(dtype, CategoricalDtype):
        col = col.astype(type(col.values[0]))
        out = get_dtype(col)
    elif np.issubdtype(dtype, np.floating):
        out = 'float32'
    elif np.issubdtype(dtype, np.integer):
        if col.max() < 32767:
            out = 'int16'
        else:
            out = 'int32'
    elif np.issubdtype(dtype, np.object_):
        size = int(col.astype(str).str.len().max())
        out = 'S{:}'.format(size)
    else:
        out = dtype

    return out


def to_records_array(df):
    """
    Convert pandas DataFrame to numpy Records Array

    Parameters
    ----------
    df : pandas.DataFrame
        Pandas DataFrame to be converted

    Returns
    -------
    numpy.rec.array
        Records array of input df
    """
    meta_arrays = []
    dtypes = []
    for c_name, c_data in df.items():
        dtype = get_dtype(c_data)

        if np.issubdtype(dtype, np.bytes_):
            data = c_data.astype(str).str.encode('utf-8').values
        else:
            data = c_data.values

        arr = np.array(data, dtype=dtype)
        meta_arrays.append(arr)
        dtypes.append((c_name, dtype))

    return np.core.records.fromarrays(meta_arrays, dtype=dtypes)


def row_col_indices(sc_point_gids, row_length):
    """
    Convert supply curve point gids to row and col indices given row length

    Parameters
    ----------
    sc_point_gids : int | list | ndarray
        Supply curve point gid or list/array of gids
    row_length : int
        row length (shape[1])

    Returns
    -------
    row : int | list | ndarray
        row indices
    col : int | list | ndarray
        row indices
    """
    rows = sc_point_gids // row_length
    cols = sc_point_gids % row_length

    return rows, cols


def unstupify_path(path):
    """
    Utility to create sensical os agnostic paths from relative or local path
    such as:
    - ~/file
    - file
    - /.
    - ./file

    Parameters
    ----------
    path : str
        Path or relative path

    Returns
    -------
    path: str
        Absolute/real path
    """
    path = os.path.expanduser(path)
    if not os.path.isabs(path) and not path.startswith('/'):
        path = os.path.realpath(path)

    return path


def write_json(path, data):
    """
    Write data to given json file

    Parameters
    ----------
    path : str
        Path to .json file to save data too
    data : dict
        Data to save to json file at path
    """
    assert path.endswith('.json'), "path should be to a .json file"
    with open(path, 'w', encoding='utf-8') as f:
        json.dump(data, f, ensure_ascii=False, indent=4)


def pd_date_range(*args, **kwargs):
    """A simple wrapper on the pd.date_range() method that handles the closed
    vs. inclusive kwarg change in pd 1.4.0"""
    incl = version.parse(pd.__version__) >= version.parse('1.4.0')

    if incl and 'closed' in kwargs:
        kwargs['inclusive'] = kwargs.pop('closed')
    elif not incl and 'inclusive' in kwargs:
        kwargs['closed'] = kwargs.pop('inclusive')
        if kwargs['closed'] == 'both':
            kwargs['closed'] = None

    return pd.date_range(*args, **kwargs)



# -*- coding: utf-8 -*-
# pylint: disable=all
"""
Multi Year ResourceX Command Line Interface
"""
import click
import logging
import os

from rex.resource_extraction.resource_cli import box as box_cmd
from rex.resource_extraction.resource_cli import dataset as dataset_grp
from rex.resource_extraction.resource_cli import multi_site as multi_site_cmd
from rex.resource_extraction.resource_cli import region as region_cmd
from rex.resource_extraction.resource_cli import site as site_cmd
from rex.resource_extraction.resource_extraction import (MultiYearResourceX,
                                                         MultiYearNSRDBX,
                                                         MultiYearWindX,
                                                         MultiYearWaveX)
from rex.utilities.cli_dtypes import STRLIST, INTLIST
from rex.utilities.loggers import init_logger
from rex import __version__

logger = logging.getLogger(__name__)


@click.group()
@click.version_option(version=__version__)
@click.option('--resource_path', '-h5', required=True,
              type=click.Path(),
              help=('Path to Resource .h5 files'))
@click.option('--out_dir', '-o', required=True, type=click.Path(),
              help='Directory to dump output files')
@click.option('--years', '-yrs', type=INTLIST, default=None,
              show_default=True,
              help='List of years to access, by default None')
@click.option('--hsds', '-hsds', is_flag=True,
              help=("Boolean flag to use h5pyd to handle .h5 'files' hosted "
                    "on AWS behind HSDS"))
@click.option('--res_cls', '-res',
              type=click.Choice(['Resource', 'NSRDB', 'Wind', 'Wave'],
                                case_sensitive=False),
              default='Resource', show_default=True,
              help='Resource type')
@click.option('--log_file', '-log', default=None, type=click.Path(),
              show_default=True,
              help='Path to .log file, if None only log to stdout')
@click.option('-v', '--verbose', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def main(ctx, resource_path, out_dir, years, hsds, res_cls, log_file,
         verbose):
    """
    Multi-year ResourceX Command Line Interface
    """
    ctx.ensure_object(dict)
    ctx.obj['H5'] = resource_path
    ctx.obj['OUT_DIR'] = out_dir
    ctx.obj['CLS_KWARGS'] = {'years': years, 'hsds': hsds}

    if res_cls == 'Resource':
        ctx.obj['CLS'] = MultiYearResourceX
    elif res_cls == 'NSRDB':
        ctx.obj['CLS'] = MultiYearNSRDBX
    elif res_cls == 'Wind':
        ctx.obj['CLS'] = MultiYearWindX
    elif res_cls == 'Wave':
        ctx.obj['CLS'] = MultiYearWaveX

    if verbose:
        log_level = 'DEBUG'
    else:
        log_level = 'INFO'

    if log_file is not None:
        log_dir = os.path.dirname(log_file)
        if not os.path.exists(log_dir):
            os.makedirs(log_dir)

    init_logger('rex', log_file=log_file, log_level=log_level)

    logger.info('Extracting Resource data from {}'.format(resource_path))
    logger.info('Outputs to be stored in: {}'.format(out_dir))


@main.group()
@click.option('--dataset', '-d', type=str, required=True,
              help='Dataset to extract')
@click.pass_context
def dataset(ctx, dataset):
    """
    Extract a single dataset
    """
    ctx.invoke(dataset_grp, dataset=dataset)


@dataset.command()
@click.option('--lat_lon', '-ll', nargs=2, type=float,
              default=None, show_default=True,
              help='(lat, lon) coordinates of interest')
@click.option('--gid', '-gid', type=int, default=None, show_default=True,
              help='Resource gid of interest')
@click.pass_context
def site(ctx, lat_lon, gid):
    """
    Extract the nearest pixel to the given (lat, lon) coordinates
    OR the given resource gid
    """
    ctx.invoke(site_cmd, lat_lon=lat_lon, gid=gid)


@dataset.command()
@click.option('--region', '-r', type=str, required=True,
              help='Region to extract')
@click.option('--region_col', '-col', type=str, default='state',
              show_default=True,
              help='Meta column to search for region')
@click.option('--timestep', '-ts', type=str, default=None,
              show_default=True,
              help='Timestep to extract')
@click.pass_context
def region(ctx, region, region_col, timestep):
    """
    Extract a single dataset for all gids in the given region
    """
    ctx.invoke(region_cmd, region=region, region_col=region_col,
               timestep=timestep)


@dataset.command()
@click.option('--lat_lon_1', '-ll1', nargs=2, type=float,
              required=True,
              help='One corner of the bounding box')
@click.option('--lat_lon_2', '-ll2', nargs=2, type=float,
              required=True,
              help='The other corner of the bounding box')
@click.option('--file_suffix', '-fs', default=None,
              show_default=True,
              help='File name suffix')
@click.option('--timestep', '-ts', type=str, default=None,
              show_default=True,
              help='Timestep to extract')
@click.pass_context
def box(ctx, lat_lon_1, lat_lon_2, file_suffix, timestep):
    """
    Extract all pixels in the given bounding box
    """
    ctx.invoke(box_cmd, lat_lon_1=lat_lon_1, lat_lon_2=lat_lon_2,
               file_suffix=file_suffix, timestep=timestep)


@dataset.command()
@click.option('--sites', '-s', type=click.Path(exists=True), required=True,
              help=('.csv or .json file with columns "latitude", "longitude" '
                    'OR "gid"'))
@click.pass_context
def multi_site(ctx, sites):
    """
    Extract multiple sites given in '--sites' .csv or .json as
    "latitude", "longitude" pairs OR "gid"s
    """
    ctx.invoke(multi_site_cmd, sites=sites)


@dataset.group(invoke_without_command=True)
@click.option('--region', '-r', type=str, default=None,
              show_default=True,
              help='Region to extract')
@click.option('--region_col', '-col', type=str, default='state',
              show_default=True,
              help='Meta column to search for region')
def map_means(ctx, region, region_col):
    """
    Map temporal means for given dataset in region if given.
    """
    logger = ctx.obj["LOGGER"]
    if ctx.invoked_subcommand is None:
        with ctx.obj['CLS'](ctx.obj['H5'], **ctx.obj['CLS_KWARGS']) as f:
            map_df = f.get_means_map(ctx.obj["DATASET"], year=None,
                                     region=region,
                                     region_col=region_col)

        out_path = "{}-means.csv".format(dataset)
        out_path = os.path.join(ctx.obj['OUT_DIR'], out_path)
        logger.info('Saving data to {}'.format(out_path))
        map_df.to_csv(out_path)
    else:
        ctx.obj['REGION'] = region
        ctx.obj['REGION_COL'] = region_col


@map_means.command()
@click.option('--year', '-yr', type=str, required=True,
              help='Year to average')
@click.pass_context
def year(ctx, year):
    """
    Map means for a given year
    """
    logger = ctx.obj["LOGGER"]
    with ctx.obj['CLS'](ctx.obj['H5'], **ctx.obj['CLS_KWARGS']) as f:
        map_df = f.get_means_map(ctx.obj["DATASET"], year=year,
                                 region=ctx.obj['REGION'],
                                 region_col=ctx.obj["REGION_COL"])

    out_path = "{}-{}.csv".format(dataset, year)
    out_path = os.path.join(ctx.obj['OUT_DIR'], out_path)
    logger.info('Saving data to {}'.format(out_path))
    map_df.to_csv(out_path)


@map_means.command()
@click.option('--years', '-yrs', type=STRLIST, required=True,
              help='List of Years to average')
@click.pass_context
def multi_year(ctx, years):
    """
    Map means for a given subset of years
    """
    logger = ctx.obj["LOGGER"]
    with ctx.obj['CLS'](ctx.obj['H5'], **ctx.obj['CLS_KWARGS']) as f:
        map_df = f.get_means_map(ctx.obj["DATASET"], year=years,
                                 region=ctx.obj['REGION'],
                                 region_col=ctx.obj["REGION_COL"])

    out_path = "{}_{}-{}.csv".format(dataset, years[0], years[-1])
    out_path = os.path.join(ctx.obj['OUT_DIR'], out_path)
    logger.info('Saving data to {}'.format(out_path))
    map_df.to_csv(out_path)


if __name__ == '__main__':
    try:
        main(obj={})
    except Exception:
        logger.exception('Error running MultiYearResourceX CLI')
        raise



# -*- coding: utf-8 -*-
"""
.h5 rechunking tool
"""
from .chunk_size import ArrayChunkSize, TimeseriesChunkSize
from .combine_h5 import CombineH5
from .rechunk_h5 import RechunkH5, get_dataset_attributes
